#!ruby

# """Hypocycloid cam generator
# Generate dxfs of hypocycloid cams for cycloid drives

# Copyright 2009, Alex Lait
# Version 	v0.2 (09/13/09)
# License 	GPL
# Homepage 	http://www.zincland.com/hypocycloid

# Credit to:
	# Formulas to describe a hypocycloid cam
	# http://gears.ru/transmis/zaprogramata/2.139.pdf

	# Insperational thread on CNCzone
	# http://www.cnczone.com/forums/showthread.php?t=72261

	# Documenting and updating the sdxf library
	# http://www.kellbot.com/sdxf-python-library-for-dxf/

	# Formulas for calculating the pressure angle and finding the limit circles
	# http://imtuoradea.ro/auo.fmte/files-2007/MECATRONICA_files/Anamaria_Dascalescu_1.pdf

# Notes:
	# Does not currently do ANY checking for sane input values and it
	# is possible to create un-machinable cams, use at your own risk

	# Suggestions:
	# - Eccentricity should not be more than the roller radius
	# - Has not been tested with negative values, may have interesting results :)
# """

# import getopt # python libs
# import sys
# import sdxf

module Mixin

  module Hypocycloid

    math = Math # a local pointer to the Math lib
    
    ### VARIABLE DEFAULTS
    #
      p   = 0.00  # Tooth Pitch
      d   = 0.00  # Roller Diameter
      e   = 0.00  # Eccentricity
      n   = 4     # Number of Teeth in Cam (min 4)
      s   = 8     # Line segements in Cam (cannot be less than n*10)
      b   =-1.00  # pin Bolt circle diameter
      f   = "xxx" # output Filename
      ang = 50.0  # Pressure Angle Limit
      c   = 0.00  # offset in pressure angle
      x   = 0.00  # 
      y   = 0.00  # 
      i   = 0     # 

    #
    ###

    def command_line_usage()
      #
      puts("Useage:\n")
      #
      puts('-p = Tooth Pitch              (float)')
      puts('-b = Pin bolt circle diameter (float)')
      puts('     -b overrides -p')
      puts('-d = Roller Diameter          (float)')
      puts('-e = Eccentricity             (float)')
      puts('-a = Pressure angle limit     (float)')
      puts('-c = offset in pressure angle (float)')
      puts('-n = Number of Teeth in Cam   (integer)')
      puts('-s = Line segements in dxf    (integer)')
      puts('-f = output filename          (string)')
      puts(' ')
      puts('-h = Help (Exclusive use. If present,')
      puts('      all other arguments are ignored.)')
      #
      puts("\nExample:")
      puts('> hypocycloid.rb -p 0.08 -d 0.15 -e 0.05 -a 50.0 -c 0.01 -n 10 -s 400 -f foo.dxf')
    end
    
    ### Proxy class
    #
    class << self 

      #  toPolar( x, y )
      #
      #
      def toPolar(x, y)
        return (x**2 + y**2)**0.5, math.atan2(y, x)
      end

      #  toRect( r, a )
      #
      #
      def toRect(r, a)
        return r*math.cos(a), r*math.sin(a)       
      end

      #  calcyp( a, e, n )
      #
      #
      def calcyp(a,e,n)
        return math.atan(math.sin(n*a)/(math.cos(n*a)+(n*p)/(e*(n+1))))
      end

      #  calcX( p, d, e, n, a )
      #
      #
      def calcX(p,d,e,n,a)
        return (n*p)*math.cos(a)+e*math.cos((n+1)*a)-d/2*math.cos(calcyp(a,e,n)+a)
      end

      #  calcY( p, d, e, n, a )
      #
      #
      def calcY(p,d,e,n,a)
        return (n*p)*math.sin(a)+e*math.sin((n+1)*a)-d/2*math.sin(calcyp(a,e,n)+a)
      end

      #  calcPressureAngle( p, d, n, a )
      #
      #
      def calcPressureAngle(p,d,n,a)
        ex = 2**0.5
        r3 = p*n
        rg = r3/ex
        pp = rg * (ex**2 + 1 - 2*ex*math.cos(a))**0.5 - d/2
        return math.asin( (r3*math.cos(a)-rg)/(pp+d/2))*180/math.pi
      end

      #  calcPressureLimit( p, d, e, n, a )
      #
      #
      def calcPressureLimit(p,d,e,n,a)
        ex = 2**0.5
        r3 = p*n
        rg = r3/ex
        q = (r3**2 + rg**2 - 2*r3*rg*math.cos(a))**0.5
        x = rg - e + (q-d/2)*(r3*math.cos(a)-rg)/q
        y = (q-d/2)*r3*math.sin(a)/q
        return (x**2 + y**2)**0.5
      end

      #  checkLimit( x, y, maxrad, minrad, offset )
      #
      #
      def checkLimit(x,y,maxrad,minrad,offset)
        r, a = toPolar(x, y)
        if (r > maxrad) or (r < minrad):
          r = r - offset
          x, y = toRect(r, a)
        return x, y
      end

    end
    #
    ### end Proxy class

    if $0 != "SketchUp" || $0 == File.basename(__FILE__)
      #
      # Only process command line args if run from command line
      # in normal Ruby. (Not in embedded Ruby under SketchUp.)
      #
      #require(File.join(File.dirname(__FILE__),'hypocycloid_cli'))
      #
      if ARGV.empty?
        #
        #  No command line args were passed !
        #
        puts("Error running file: '#{__FILE__}' #<RuntimeError: No command line args were passed!>\n")
        #
        #
        Mixin::Hypocycloid.CLI.output__usage()
        #
        exit(2)
        #
      else
        if ARGV[0].any?{|arg| arg =~ /(-h|-H)/ }  # output the argument help
          #
          Mixin::Hypocycloid.command_line_usage()
          #
          exit(0)
          #
        else
          #
          # init default values
          #
          p = 0.00
          d = 0.00
          e = 0.00
          n = 0.00
          s = 0.00
          b =-1.00
          f = "xxx"
          ang = 50.0
          c = 0.00
          x = 0.00
          y = 0.00
          i = 0
          #
          # Process the arguments
          #
          begin
            #
            argv = Hash[*ARGV] # convert the ARGV array to a hash
            #
          rescue ArgumentError => e
            #
            if e.message.include?(': odd number of arguments for Hash')
              puts("Error running file: '#{__FILE__}' #<ArgumentError: odd number of arguments were passed!>\n")
            else
              puts("Error running file: '#{__FILE__}' #<#{e.class.name}: #{e.message}>\n")
            end
            puts(e.backtrace)
            #
            exit(2)
            #
          rescue => e
            #
            puts("Error running file: '#{__FILE__}' #<#{e.class.name}: #{e.message}>\n")
            puts(e.backtrace)
            #
            exit(2)
            #
          else
            #
            for opt,val in argv
              case opt
              when '-p', '-P'
                p = Float(val)
              when '-b', '-B'
                b = Float(val)
              when '-d', '-D'
                d = Float(val)
              when '-e', '-E'
                e = Float(val)
              when '-n', '-N'
                n = Integer(val)
              when '-s', '-S'
                s = Integer(val)
              when '-a', '-A'
                ang = Float(val)
              when '-c', '-C'
                c = Float(val)
              when '-f', '-F'
                f = val
              else
                puts("Warning: unhandled option #{opt} = '#{val}'")
              end # case
            end # for
            #
          end # Processing arguments
          #
          # if -b was specifed, calculate the tooth pitch for use in cam generation
          p = b/n if b > 0
          q = 2*math.pi/Float(s)
          
          #
          exit(0)
          #
        end # if help arg specified
        #
      end # if ARGV.empty?
      #
    end # if $0 != "SketchUp"

    #DXF = Class.new(File)
    #dxf = DXF.new( f )
    dxf=sdxf.Drawing()
    
    # test for and create layers if needed
    #dxf.layers.append(sdxf.Layer(name="text", color=2) )	#yellow text layer
    dxf.layers.append(sdxf.Layer(name="cam", color=1) ) 	#red cam layer
    dxf.layers.append(sdxf.Layer(name="roller", color=5) )	#blue roller layer
    dxf.layers.append(sdxf.Layer(name="pressure", color=3) )	#pressure angle limit layer

    # add attribute dictionary to component after creation
    # Add text to discribe the cam - autodocument
    dxf.append( sdxf.Text( "pitch="       +str(p), point=(p*n+d, 0.7), layer="text", height=0.1) )
    dxf.append( sdxf.Text( "pin diameter="+str(d), point=(p*n+d, 0.5), layer="text", height=0.1) )
    dxf.append( sdxf.Text( "eccentricity="+str(e), point=(p*n+d, 0.3), layer="text", height=0.1) )
    dxf.append( sdxf.Text( "# of teeth="  +str(n), point=(p*n+d, 0.1), layer="text", height=0.1) )
    dxf.append( sdxf.Text( "pressure angle limit="  +str(ang), point=(p*n+d,-0.1), layer="text", height=0.1) )
    dxf.append( sdxf.Text( "pressure angle offset="  +str(c), point=(p*n+d,-0.3), layer="text", height=0.1) )

    # Find the pressure angle limit circles
    minAngle = -1.0
    maxAngle = -1.0
    for i in range(0, 180):
            x = calcPressureAngle(p,d,n,float(i)*math.pi/180)
            minAngle = float(i) if( (x < ang) and (minAngle < 0) )        
            maxAngle = float(i-1) if( (x < -ang) and (maxAngle < 0) )
    end
    dxf.append( sdxf.Text( "min angle="  +str(minAngle), point=(p*n+d,-0.5), layer="text", height=0.1) )
    dxf.append( sdxf.Text( "max angle="  +str(maxAngle), point=(p*n+d,-0.7), layer="text", height=0.1) )
    minRadius = calcPressureLimit(p,d,e,n,minAngle*math.pi/180)
    maxRadius = calcPressureLimit(p,d,e,n,maxAngle*math.pi/180)
    dxf.append( sdxf.Circle(center=(-e, 0), radius=minRadius, layer="pressure") )
    dxf.append( sdxf.Circle(center=(-e, 0), radius=maxRadius, layer="pressure") )

    #generate the cam profile - note: shifted in -x by eccentricicy amount
    i=0
    x1 = calcX(p,d,e,n,q*i)
    y1 = calcY(p,d,e,n,q*i)
    x1, y1 = checkLimit(x1,y1,maxRadius, minRadius, c)
    x0, y0 = x1-e, y1 # added by Dan Rathbun
    for i in range(0, s): # should this be s-1 so segs is even ?
            x2 = calcX(p,d,e,n,q*(i+1))
            y2 = calcY(p,d,e,n,q*(i+1))
            x2, y2 = checkLimit(x2,y2,maxRadius, minRadius, c)
            # test for last seg, make last end pt be the start pt
      dxf.append( sdxf.Line(points=[(x1-e,y1),  (x2-e,y2)], layer="cam" ) )
            x1 = x2
            y1 = y2
    end
    #add a circle in the center of the cam
    dxf.append( sdxf.Circle(center=(-e, 0), radius=d/2, layer="cam") )

    #generate the pin locations
    for i in range(0, n+1):
      x = p*n*math.cos(2*math.pi/(n+1)*i)
      y = p*n*math.sin(2*math.pi/(n+1)*i)
      dxf.append( sdxf.Circle(center=(x,y), radius=d/2, layer="roller") )
    end
    
    #add a circle in the center of the pins
    dxf.append( sdxf.Circle(center=(0, 0), radius=d/2, layer="roller") )


    begin
      dxf.saveas(f)
    rescue
      print("Problem saving file")
      exit(2)
    end

  end # module Hypocycloid

end # outer module


# hypocycloid_cli.rb
#
#
module Mixin::Hypocycloid

  module CLI

    math = Math # a local pointer to the Math lib
    
    ###
    #
      p = 0.00
      d = 0.00
      e = 0.00
      n = 0.00
      s = 0.00
      b =-1.00
      f = "xxx"
      ang = 50.0
      c = 0.00
      x = 0.00
      y = 0.00
      i = 0

    #
    ###

    def self.output_usage()
      #
      puts("Usage:\n")
      #
      puts('-p = Tooth Pitch              (float)')
      puts('-b = Pin bolt circle diameter (float)')
      puts('     -b overrides -p')
      puts('-d = Roller Diameter          (float)')
      puts('-e = Eccentricity             (float)')
      puts('-a = Pressure angle limit     (float)')
      puts('-c = offset in pressure angle (float)')
      puts('-n = Number of Teeth in Cam   (integer)')
      puts('-s = Line segements in dxf    (integer)')
      puts('-f = output filename          (string)')
      puts(' ')
      puts('-h = Help (Exclusive use. If present,')
      puts('      all other arguments are ignored.)')
      #
      puts("\nExample:")
      puts('> hypocycloid.rb -p 0.08 -d 0.15 -e 0.05 -a 50.0 -c 0.01 -n 10 -s 400 -f foo.dxf')
    end

  end # module CLI

end # outer modules
